{#
#  Polfatt 5
#  Copyright 2012 by Marco Spadini
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#
### ################### README ####################################################
### DESCRIPTION:                                                                ###
###    This is a port to GNU Pascal, with slight modifications, of a program I  ###
### wrote in 1993 in TurboPascal as a test. Its purpose is to factorize in Q[X] ###
### (actually in Z[X]), polynomials with integer (small) coefficients.          ###
###    Unlike serious and powerful program based on commutative alebra tools,   ###
### This program is based on the Schubert-Kronecker algorithm which is not very ###
### efficient, actually, I think it has an exponential complexity... not good   ### 
### at all for serious factoring. However, for polynomials with rather small    ###
### coefficients and low degree (let us say smaller than 7 or 8) I find it is   ###
### fast and reliable. (One should also beware of factors that have rational    ###
### cofficients with large numerators or denominators.)                         ###
###   In fact, there is a limitation built in this program (not in the          ###
### underlying algorithm, only in the implementation): it cannot correctly      ###
### treat irreducible factors of degree higher than 5; so, for instance, if a   ###
### polynomial P of degree 12 has two irreducible factors of degree 6, the      ###
### program will incorrectly state that P is irreducible (it would also take a  ###
### _long_ time to say so). For this reason, NEVER use this program to          ###
### factorize anything of degree larger than 11 (although it might be able to   ###
### factorize it: an answer with factors of degree at most five is certainly    ###
### correct). This strange limitation is due to the  nasty nested cycle in      ###
### function calnodi. I could have used some cleverer way  to allow for         ###
### arbitrary degree but, as I said, this was meant to be a test and  neither   ###
### then I felt, nor now I feel, inclined to waste time on a test (one  should  ###
### bear in mind the exponential complexity of the algorithm which make  rather ###
### pointless to implement a mechanism for arbitrarily large degree). What      ###
### is more, now (for me) this is merely legacy code (but fun to look at!).     ###
###                                                                             ###
### One more caveat. Beware that the program wants integer coefficients for the ###
### input polynomial but quietly accepts other stuff. If you feed unexpected    ###
### data you may get what you deserve: unpredictable behaviour!                 ###
###                                                                             ###
###   RUNNING the program:                                                      ###
###                                                                             ###
###      This is source code written in GNU Pascal. To run the program obtain   ###
###      the compiler (gpc under Linux) compile the source file and launch the  ###
###      object according to the rules of your system. For instance, in my      ###
###      Linux box I run:                                                       ###
###      gpc POLFATT5.pas -o polfatt                                            ###
###      ./polfatt                                                              ###
###                                                                             ###
###       Have fun!                                                             ###
###                                                                             ###
###                                                                             ###
###       For any issues concerning this program please contact the author:     ###
###       Dr. Marco Spadini                                                     ###
###       Dipartimento di Matematica e Informatica,                             ###
###       Via S. Marta 3, 50139 Firenze,                                        ###
###       ITALY                                                                 ###
###       web: http:\\www.dma.unifi.it\~spadini                                 ###
###       e-mail: marco.spadini@math.unifi.it                                   ###
###                                                                             ###
###                                                                             ###
###################################################################################
#}

program scomposizione; { original version: 5 -18/4/'93- port: 6/12/2012 }

 type
     ratio = record
                 n:longint;
                 d:longint;
             end;
       pol = array[0..11] of ratio;
      cpol = record
                 g:integer;
                 l:pol;
             end;
     divis = array[0..200] of longint;
      nodi = array[0..5] of divis;
     nnodi = array[0..5] of integer;
   elirrid = array[1..12] of cpol;

  const
    nul  : ratio = (n:0;d:1);
    zero : cpol = (g:0;l:((n:0;d:1),(n:0;d:1),(n:0;d:1),(n:0;d:1),(n:0;d:1),
                 (n:0;d:1),(n:0;d:1),(n:0;d:1),(n:0;d:1),(n:0;d:1),(n:0;d:1),
                 (n:0;d:1)));
    uno  : cpol = (g:0;l:((n:1;d:1),(n:0;d:1),(n:0;d:1),(n:0;d:1),(n:0;d:1),
                 (n:0;d:1),(n:1;d:1),(n:0;d:1),(n:0;d:1),(n:0;d:1),(n:0;d:1),
                 (n:0;d:1)));

 var p        :cpol;
     f        :elirrid;
     cont,t   :integer;
     r        :char;
     tm1,tm2  :real;


 function mcd(a,b:longint):longint;
   var s,r,md :longint;
   begin
     md:=a;
     s:=b;
     while s<>0 do
       begin
         r:=md MOD abs(s);     
         md:=s;
         s:=r
       end;
     mcd:=md;
   end;

 procedure riduci(var a:ratio);
   var b  :longint;
   begin
     b:=mcd(a.n,a.d);
     a.n:=a.n div b;
     a.d:=a.d div b;
   end;

 procedure sommaq(a,b:ratio; var c:ratio);
   var dn,a1  :longint;
   begin
     dn:=mcd(a.d,b.d);
     a1:=a.d div dn;
     c.d:=a1*b.d;
     c.n:=(a1*b.n+a.n*(b.d div dn));
     riduci(c);
   end;

 procedure moltip(a,b:ratio; var c:ratio);
   begin
     c.n:=a.n*b.n;
     c.d:=a.d*b.d;
     riduci(c);
   end;

 procedure recip(var a:ratio);
   var b  :longint;
   begin
     b:=a.n;
     a.n:=a.d;
     a.d:=b;
   end;

 procedure opposto(var a:ratio);
   begin
     a.n:=-a.n;
   end;

 procedure divid(a,b:ratio; var c:ratio);
   begin
     recip(b);
     moltip(a,b,c);
   end;

 procedure sottrai(a,b:ratio; var c:ratio);
   begin
     opposto(b);
     sommaq(a,b,c);
   end;

 function signum(r:ratio):integer;
   var s  :boolean;
   begin
     s:=((r.n>0) and (r.d>0)) or ((r.n<0) and (r.d<0));
     if s then
       signum:=1
     else
       if (r.n<>0) then
         signum:=-1
       else
         signum:=0;
    end;

 function deg(p:pol):integer;
   var t:integer;
   begin
     deg:=0;
     for t:=1 to 11 do if (p[t].n<>0) then deg:=t;
   end;

 procedure spoly(a,b:cpol;var p:cpol);
   var t,k  :integer;
   begin
     p:=zero;
     k:=a.g;
     if (k<b.g) then k:=b.g;
     for t:=0 to k do sommaq(a.l[t],b.l[t],p.l[t]);
     p.g:=deg(p.l);
   end;

 procedure ppoly(a,b:cpol;var p:cpol);
   var t,s,k  :integer;
       r1,r2  :ratio;
   begin
     p:=zero;
     k:=a.g+b.g;
     for t:=k downto 0 do
       begin
         r2:=NUL;
         for s:=0 to t do
           begin
             moltip(a.l[s],b.l[t-s],r1);
             sommaq(r1,r2,r2);
           end;
         p.l[t]:=r2;
         p.g:=deg(p.l);
       end;
   end;

 procedure mspoly(r:longint; var p:cpol);
   var t  :integer;
       r1 :ratio;
   begin
     r1.n:=r;
     r1.d:=1;
     for t:=0 to p.g do moltip(p.l[t],r1,p.l[t]);
   end;

 procedure lagrange(m:integer;v:pol;var pl:cpol);
   var i,j     :integer;
       p,q     :cpol;
   begin
     pl:=zero;
     q:=zero;
     q.g:=1;
     q.l[1].n:=1;
     for i:=0 to m do
       begin
         p:=uno;
         for j:=0 to m do if (i<>j) then
                            begin
                              q.l[1].d:=v[i].n-v[j].n;
                              q.l[0].n:=-v[j].n;
                              q.l[0].d:=q.l[1].d;
                              riduci(q.l[0]);
                              ppoly(q,p,p);
                            end;
         mspoly(v[i].d,p);
         spoly(p,pl,pl);
       end;
   end;

 function evintpoly(p:cpol;x:longint):longint;
   var t    :integer;
       y    :longint;
   begin
     y:=p.l[p.g].n;
     for t:=p.g-1 downto 0 do y:=y*x+p.l[t].n;
     evintpoly:=y;
   end;

 procedure dintpoly(p1,p2:cpol;var q,r:cpol);
   var t,k  :integer;
       flag :boolean;
   begin
     flag:=TRUE;
     q:=zero;
     r:=p1;
     while ((r.g>=p2.g) and flag) do
       begin
         t:=r.g-p2.g;
         if ((r.l[r.g].n MOD abs(p2.l[p2.g].n))=0) then   {modifica}
           begin
             q.l[t].n:=r.l[r.g].n div p2.l[p2.g].n;
             for k:=p2.g downto 0 do r.l[k+t].n:=r.l[k+t].n-q.l[t].n*p2.l[k].n;
           end
         else flag:=FALSE;
         r.g:=deg(r.l);
       end;
     q.g:=p1.g-p2.g;
   end;

 procedure divisori(n:longint; var f:divis; var nd:integer);
   var t,m :longint;
       h   :integer;
   begin
     n:=abs(n);
     m:=n div 2;
     nd:=0;
     f[0]:=1;
     for t:=2 to m do
       begin
         if ((n mod t)=0) then
           begin
             nd:=nd+1;
             f[nd]:=t;
           end;
       end;
     nd:=nd+1;
     f[nd]:=n;
     for h:=0 to nd do f[nd+h+1]:=-f[h];
     nd:=nd*2+1;
   end;

 procedure linhunter(var p:cpol);
   var s,t               :longint;
       den,quoz,resto,q  :cpol;
       f0,f1             :divis;
       i0,i1,h,k         :integer;
       flag              :boolean;
   begin
     while (p.l[0].n=0) do
       begin
         q:=zero;
         cont:=cont+1;
         f[cont].g:=1;
         f[cont].l[1].n:=1;
         q.g:=p.g-1;
         for t:=0 to q.g do q.l[t]:=p.l[t+1];
         p:=q;
       end;
     repeat
       flag:=TRUE;
       den:=uno;
       den.g:=1;
       divisori(p.l[p.g].n,f1,i1);
       divisori(p.l[0].n,f0,i0);
       h:=0;
       while ((h<=i0) and flag) do
         begin
           k:=0;
           while ((k<=i1) and flag) do
             begin
               den.l[0].n:=f0[h];
               den.l[1].n:=f1[k];
               dintpoly(p,den,quoz,resto);
               if (resto.l[0].n=0) then
                 begin
                   cont:=cont+1;
                   f[cont]:=den;
                   p:=quoz;
                   flag:=FALSE;
                 end;
               k:=k+1;
             end;
           h:=h+1;
         end;
     until flag;
   end;

 function maxcden(p:cpol):longint;
   var maxd  :longint;
       t     :integer;
   begin
      t:=0;
      repeat
        maxd:=p.l[t].n;
        t:=t+1;
      until ((maxd<>0) or (t>5));
      for t:=0 to p.g do
        if (p.l[t].n<>0) then maxd:=mcd(maxd,p.l[t].n);
      maxcden:=maxd;
   end;

 procedure gendivnl(p:cpol);
   var m                    :nodi;
       i                    :nnodi;
       r                    :pol;
       quoz,resto,den       :cpol;
       t                    :longint;
       i0,i1,i2,i3,i4,
       i5,numnod            :integer;
       flag                 :boolean;

   function polint(p:cpol):boolean;
     var flag  :boolean;
         t     :integer;
     begin
       flag:=TRUE;
       t:=0;
       while ((t<=p.g) and flag) do
         begin
           flag:=flag and (p.l[t].d=1);
           t:=t+1;
         end;
       polint:=flag;
     end;

   procedure calnodi(var m:nodi; var i:nnodi);
     var h            :integer;
         t,nosx,nod   :longint;
     begin
       for h:=0 to 5 do i[h]:=0;
       nosx:=-(p.g div 4);
       numnod:=p.g div 2;
       for t:=0 to numnod do
         begin
           nod:=nosx+t;
           divisori(evintpoly(p,nod),m[t],i[t]);
           r[t].n:=nod;
         end;
       i[0]:=(i[0]+1) div 2;
     end;

   begin
     repeat
     flag:=TRUE;
     calnodi(m,i);
     i0:=0;
     while ((i0<=i[0]) and (flag)) do
       begin
         i1:=0;
         while ((i1<=i[1]) and (flag)) do
           begin
             i2:=0;
             while ((i2<=i[2]) and (flag)) do
               begin
                 i3:=0;
                 while ((i3<=i[3]) and (flag)) do
                   begin
                     i4:=0;
                     while ((i4<=i[4]) and (flag)) do
                       begin
                         i5:=0;
                         while ((i5<=i[5]) and (flag)) do
                           begin
                             r[0].d:=m[0][i0];
                             r[1].d:=m[1][i1];
                             r[2].d:=m[2][i2];
                             r[3].d:=m[3][i3];
                             r[4].d:=m[4][i4];
                             r[5].d:=m[5][i5];
                             lagrange(numnod,r,den);
                             if (den.g>1) then
                               begin
                                 if (polint(den)) then
                                   begin
                                     if (abs(maxcden(den))=1) then
                                       begin
                                         dintpoly(p,den,quoz,resto);
                                         if ((resto.g=0) and (resto.l[0].n=0)) then
                                         begin
                                           gendivnl(den);
                                           p:=quoz;
                                           flag:=FALSE;
                                         end;
                                       end;
                                   end;
                               end;
                             i5:=i5+1
                           end;
                           i4:=i4+1;
                         end;
                       i3:=i3+1;
                     end;
                   i2:=i2+1;
                 end;
               i1:=i1+1;
             end;
           i0:=i0+1;
         end;
     until flag;
     if (p.g>1) then
       begin
         cont:=cont+1;
         f[cont]:=p;
       end;
   end;

  procedure gendiv(p:cpol);
    var maxd   :longint;
        t      :integer;
    begin
      maxd:=abs(maxcden(p));
      if (maxd<>1) then
        begin
          cont:=1;
          f[cont].l[0].n:=maxd;
        end;
      for t:=0 to p.g do
        p.l[t].n:=p.l[t].n div maxd;
      linhunter(p);
      if (p.g>1) then gendivnl(p);
    end;

 procedure carica(var p:cpol);
   var t  :integer;
   begin
     P:=ZERO;
     write('Insert DEGREE ');
     readln(p.g);
     for t:=p.g downto 0 do
       begin
         write('insert the coefficient of X^',t,' ');
         readln(p.l[t].n);
         p.l[t].d:=1;
       end;
     p.g:=deg(p.l);
   end;

 procedure wpoly(p:cpol);
   var t  :integer;
       ab :longint;
   begin
     write('(');
     for t:=p.g downto 1 do
       if (p.l[t].n<>0) then
         begin
           if (signum(p.l[t])>0) then
             begin
               if (t<>p.g) then write(' +')
             end
           else
             write(' -');
           ab:=abs(p.l[t].n);
           if (ab<>1) then write(ab);
           if (t<>1) then
             write('X^',t)
           else
             write('X');
         end;
           if (p.l[0].n<>0) then
             begin
               if (signum(p.l[0])>0) then
                 write(' +')
               else
                 write(' -');
               write(abs(p.l[0].n));
             end;
       write(')');
   end;

 procedure normalizza(var acc:integer);
   var t,s  :integer;
  begin
    acc:=1;
    for t:=1 to cont do
      begin
        if (signum(f[t].l[f[t].g])<0) then
          for s:=f[t].g downto 0 do
            f[t].l[s].n:=f[t].l[s].n*(-1);
      end;
    if (signum(p.l[p.g])<0) then acc:=-1;
  end;

  procedure stamparis;
    var sg,t  :integer;
    begin
      {clrscr;}
      writeln;
      writeln('      Decomposition of: ');
WRITELN('-------------------------------------------------------------------');
  write('****** ');
      wpoly(p);
writeln(' ******');
WRITELN('-------------------------------------------------------------------');
      writeln;
      writeln;
      normalizza(sg);
      {write(' FACTOR n.  1: ');
      if (sg<0) then write('-');
      wpoly(f[1]);}
      if (cont = 1) then 
        begin
          writeln(' The given polynomial is irreducible.  '); 
        end
      else 
        begin
          if (sg<0) then 
            begin
              write(' FACTOR n. 1 ',': '); 
              if (sg<0) then write('-');
              wpoly(f[1]);
              writeln;
            end; 
          for t:=2 to cont do
            begin
              write(' FACTOR n. ',t:2,': ');
              wpoly(f[t]);
              writeln;
            end; 
        end;
      writeln;
    end;

 function tempo:real;
   var
       h,m,s,c       :word;
       hv,mv,sv,cv   :integer;
     mytime   :TimeStamp;
   begin
     GetTimeStamp(mytime);
     hv:=mytime.hour;
     mv:=mytime.minute;
     sv:=mytime.second;
     cv:=mytime.microsecond;
     tempo:=3600*hv+60*mv+sv+cv/1000000;
   end;

 begin
   writeln(' Polfatt5 Copyright (C) 2012  Marco Spadini ');
   writeln;
   writeln('  This program comes with ABSOLUTELY NO WARRANTY');
   writeln(' This program is distributed in the hope that it will be useful,');
   writeln(' but WITHOUT ANY WARRANTY; without even the implied warranty of ');
   writeln(' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the ' );
   writeln(' GNU General Public License for more details.');
   writeln(' This is free software, and you are welcome to redistribute it ');
   writeln(' under certain conditions.');
   writeln;
   for t:=1 to 12 do
     f[t]:=zero;
   cont:=0;
   p:=zero;
   writeln('INSERT polynomial to be factorized ');
   carica(p);
   tm1:=tempo;
   gendiv(p);
   tm2:=tempo;
   stamparis;
   tm1:=tm2-tm1;
   writeln('**** Time: (SEC.) ',tm1:4:4,' *****');
   writeln;
 end.
